###################################
# PROLOG SECTION
# pot_shot.py
# A Turtle graphics program to play a 2-person shooting game.
#
# This game is a Python version of a game that was popular
# on mainframe computers many years ago.
#
# There are two players in this game. They take turns
# shooting projectiles at each other until one player
# hits the other player's shooting location.
#
# IMPORTANT: Both the Turtle Graphics drawing window
# and the Python Shell window need to be open and visible
# to play this game.
#
# Tested with Python 3.7.9 on Windows 10
# Fri July 16, 2021
#
# Copyright ©. All rights reserved. NAF 2021.
#
# Possible future enhancements:
# Unresolved bugs:
###################################


###################################
# ENVIRONMENT SETUP SECTION
# Imports: math, turtle, random
###################################

from turtle import *
from random import *

# We have to import the math routines this way because there is a
# radians() function in both the math and turtle modules.
import math

###################################
# FUNCTION DEFINITION SECTION
# Definitions: xcoord(), ycoord(),
#  draw_target(), declare_victory()
###################################

#
# Function to calculate the projectile x-coordinate
# at time "t" for an initial velocity "v" shot an an angle "a"
# in degrees.
#
# This formula is derived from the laws of physics for
# a moving particle subject to the law of gravity.
#
def xcoord(t,v,a):
    x=t*v*math.cos(math.radians(a))
    return x

#
# Define a function to return the y-coordinate as above.
#
def ycoord(t,v,a):
    # acceleration of gravity
    g=32.174
    y=t*v*math.sin(math.radians(a)) - (g*t*t)/2.0
    return y

#
# Define a function to draw a target.
#
# Each of the players has a shooting platform
# that is also the target to be hit by the other player.
# The targets both sit on the line at the bottom
# called the baseline.
#
def draw_target(n,turtle,baseline):
    # define length of each side of the target
    sidelen=30
    # define the colors of both targets as a tuple
    targets=("red","blue")
    
    # draw the target as a square centered
    # on the target location on the baseline.
    turtle.hideturtle()
    turtle.penup()
    # each target has a randomly chosen x-coordinate
    turtle.goto(targetx[n],baseline)
    turtle.color(targets[n])
    turtle.setheading(0)
    turtle.down()
    turtle.forward(sidelen/2)
    turtle.left(90)
    turtle.forward(sidelen)
    turtle.left(90)
    turtle.forward(sidelen)
    turtle.left(90)
    turtle.forward(sidelen)
    turtle.left(90)
    turtle.forward(sidelen/2)
    turtle.penup()
    turtle.goto(targetx[n],baseline+(sidelen/2))
    
    # label the target with its number
    turtle.write(str(n+1))

#
# Define function that declares victory for one of the players.
#
# The function will write out a victory message
# and draw a black circle on top of the destroyed
# target.
#
def declare_victory(n,mypen,baseline):
    
    mypen.hideturtle()
    mypen.up()
    mypen.goto(0,0)
    # write a victory message on the screen
    mypen.write("Player "+str(n+1)+" wins!")
    # mark the opponent's target as destroyed
    mypen.up()
    # move to the other player's spot
    mypen.goto(targetx[(n+1)%2],baseline)
    # draw a black circle on top of it
    mypen.setheading(0)
    mypen.begin_fill()
    mypen.color("black")
    mypen.circle(30)
    mypen.end_fill()
    
###################################
# PROCESSING INITIALIZATION SECTION
###################################

# set debug print flag
debug_print = False

# set the window dimensions and title
maxx=300
maxy=200
minx=-maxx
miny=-maxy
setup(width=2*maxx, height=2*maxy)
title("potshot")

# create the two bullet projectiles
# but don't draw them yet
bullet=2*[0]
for i in range(2):
    bullet[i]=Turtle()
    bullet[i].hideturtle()
    bullet[i].shape("circle")
    if i==0:
        bullet[i].color("red")
    else:
        bullet[i].color("blue")
    bullet[i].shapesize(1,1,1)
    bullet[i].penup()
 
# draw the baseline using the "mypen" drawing turtle
baseline=miny+30
mypen=Turtle()
mypen.hideturtle()
mypen.color("black")
mypen.width(5)
mypen.penup()
mypen.goto(minx,baseline)
mypen.down()
mypen.setheading(0)
mypen.showturtle()
mypen.forward(2*maxx)
mypen.hideturtle()

# create and draw the two targets
targetx=2*[0]
# the targets have random x-coordinates along the
# baseline
targetx[0]=minx + randint(30,100)
targetx[1]=maxx - randint(30,100)

# debug print
if debug_print:
    print("DEBUG: targets x-coords are", targetx)

# set the directions used by the two targets
direction=(1,-1)

# draw the two targets in the window
for i in range(2):
    # call the target drawing function
    draw_target(i,mypen,baseline)

# intialize the loop termination condition variable
endgame = False

###################################
# PROCESSING SECTION
# Loop that prompts each player for an angle and velocity
# and then shoots the projectile.
# Branching code: if/else
# Looping code: while-loop, for-loop
###################################

# keep going until the game ends
while not endgame:
    # loop through the two players as they take turns shooting
    for i in range(2):
        # set the index for the other player
        opponent = (i+1)%2
        # prompt for velocity and angle to use for next shot
        print()
        print("Player",i+1)
        velocity=float(input("velocity (10-200)? "))
        if velocity==0:
            endgame=True
            break
        angle=float(input("angle (0-90)? "))

        # move the player's projectile to its starting location
        bullet[i].hideturtle()
        bullet[i].goto(targetx[i],baseline)
        bullet[i].showturtle()
        bullet[i].down()

        # Move the projectile based on the law of gravity
        # until it falls to earth (hits the baseline) or goes
        # off the screen. The range used here and time increment
        # were determined experimentally to make the movement
        # of the projectile be smooth.
        for t in range(0,30000,10):
            tt=t*0.001
            # calculate x and y coordinates at time tt
            x=direction[i]*xcoord(tt,velocity,angle)+targetx[i]
            y=ycoord(tt,velocity,angle)+baseline
            
            # debug print
            if debug_print:
                print("DEBUG: bullet x,y is",x,y)
                
            # move the projectile along its trajectory
            bullet[i].goto(x,y)
            # stop moving the projectile if it has left the
            # allowed space in the window
            if y<baseline or x<minx or x>maxx:
                break
            # check for a win, which happens when a bullet gets
            # close enough to the opponent's target.
            #
            # the distance (20) was also determined experimentally
            if bullet[i].distance(targetx[opponent],baseline)<20:
                # write out a victory message and blacken
                # out the target
                declare_victory(i,mypen,baseline)
                # game is over, stop the loop and the game
                endgame=True
                break

        bullet[i].up()
        if endgame:
            break

###################################   
# CLEANUP, TERMINATION, AND EXIT SECTION
###################################

# wait for a mouse click to close the window
exitonclick()
