###################################
# PROLOG SECTION
# tic_tac_toe.py
#
# This program plays the traditional tic-tac-toe game,
# computer against human player.
# It uses text strings only.
#
# This game does not use optimal strategy, which would
# require additional code.
#
# It uses lists and tuples (aka arrays in other programming languages).
# Lists can be changed after they are created. Lists are delimited by
# square brackets. Tuples cannot be changed after they are created.
# Tuples are delimited by parentheses.
#
# Tested with Python 3.7.9 on Windows 10
# Fri July 16, 2021
#
# Copyright ©. All rights reserved. NAF 2021.
#
# Possible future enhancements:
# Unresolved bugs:
###################################

###################################
# ENVIRONMENT SETUP SECTION
# Imports: random
###################################

# import random number routines
from random import *

###################################
# FUNCTION DEFINITION SECTION
# Definitions: display_usage()
#  get_marker(), show_board()
#  check_for_win(),
#  pick_computer_move()
###################################

#
# Function to display the game rules and usage
#
# a usage routine is a common programming technique
#
# an alternate treatment is to have print statements
# at the beginning of the program
#
def display_usage():
    print()
    print("Welcome to the game of tic-tac-toe.")
    print("The computer will be X and you will be O.")
    print("We flip a coin to decide who goes first.")
    print()

#    
# Function to return the current value of a given board square
# as an X, O, or a blank.
#
# There are 9 cells, marked 1-9 stored in a Python list.
#
# The internal list subscripts are 0-8.
#
# The function show_board() (below) displays the board with
# the correct marker for the value.
#
def get_marker(board, n):
    # the board list (array) is defined below
    # does the cell contain an X?
    if board[n]==mark_x:
        marker='X'
    # does the cell contain an O?
    elif board[n]==mark_o:
        marker='O'
    else:
        # for an empty cell, display its number
        # the str() function converts the number
        # to a string
        marker=str(n+1)

    # this function returns a value via the
    # marker string variable
    return marker

#   
# Function to display the state of the playing board
# as a table with 9 cells in it.
#
# The function get_marker() (above) provides the
# string value to display.
#
# Note: list subscripts 0-8 represent displayed
#       values 1-9
#
def show_board(board):
    print()
    # top 3 cells
    print(' ',get_marker(board,0),'|',get_marker(board,1),'|',get_marker(board,2))
    print(" ---+---+---")
    # middle 3 cells
    print(' ',get_marker(board,3),'|',get_marker(board,4),'|',get_marker(board,5))
    print(" ---+---+---")
    # bottom 3 cells
    print(' ',get_marker(board,6),'|',get_marker(board,7),'|',get_marker(board,8))
    print()

#   
# Function to check to see if somebody has won the game.
# It looks for three in a row of any sort for computer or player.
#
# mark_test gets its value from the various calls in the
# main part of the program.
#
def check_for_win(board, mark_test):
    # debug print
    global debug_print
    if debug_print:
        print("DEBUG: enter check_for_win with", mark_test)
        
    # a win means the same mark is in three cells in a row
    win_value=3*mark_test
    # a tuples of tuples of all winning cell combinations
    # of three in a row
    wins=((0,1,2),(3,4,5),(6,7,8),(0,3,6),(1,4,7),(2,5,8),(0,4,8),(2,4,6))
    
    # Test each board cell combination to see if all the cells
    # for a given vector are occupied.
    # We loop through all the tuples of three numbers and check if
    # each of the cells has the same content.
    for vector in wins:
        if board[vector[0]]+board[vector[1]]+board[vector[2]] == win_value:
            # if all cells are the same, we have found a win
            return True
    
    # no test passed, there is no win yet
    # False is a Boolean constant
    return False

#
# Function to pick a move for the computer
#
# This function just returns a random empty cell.
#
def pick_computer_move(board, full):
    # debug print
    global debug_print
    if debug_print:
        print("DEBUG: enter pick_computer_move with", full)
    
    # To make it easier for the player to win, the computer is just
    # picking a random empty cell. A better computer strategy would
    # be for the computer to look first for ways to block
    # however, this tends to produce games ending in a draw.

    # If the main program did not check for a full board,
    # this function could loop forever
    while True and full<9:
        # try picking a random cell
        pick=randrange(1,9)
        # if it is empty, that is our pick
        if board[pick-1]==mark_empty:
           # debug print
            if debug_print:
                print("DEBUG: pick_computer_move picks", pick)
            # return our move
            return pick

    # we should never get here!
    print("There is no move possible, the board is full.")
    return 0

###################################
# PROCESSING INITIALIZATION SECTION
###################################

# set debug print flag
debug_print = False

#
# define the possible contents of squares
#

# define a free square
mark_empty=0
# define a square occupied by an X
# total of 3 Xs in a row equals 3
mark_x=1
# define a square occupied by an O
# total of 3 Os in a row equals 30
mark_o=10

# the computer is X
# in this game, the computer is always X
mark_computer = mark_x
# the player is O
# in this game, the player is always O
mark_player=mark_o

# create an empty board (a list with 9 cells,
# each of which has a value of 0).
# the value of mark_empty is 0.
board=9*[mark_empty]

# no cells are occupied yet
full_cells=0

# display rules and usage
display_usage()

# flip a coin to see who goes first
flip=randrange(1,100)
if flip>50:
    # the computer goes first
    print("The computer wins the toss and makes the first move")
    # pick a move for the computer
    computermove=pick_computer_move(board, full_cells)
    # place the mark of the computer in the chosen cell
    board[computermove-1]=mark_computer
    # increment the full cell count
    full_cells=full_cells+1
else:
    print("You win the toss and go first.")

###################################
# PROCESSING SECTION
# Loop keep alternating moves until someone wins
# or the game is a draw.
# Branching code: if/else
# Looping code: while-loop, for-loop
###################################

# keep alternating moves until someone wins or the game is a draw
while True:
   # initializes valid_move (we don't have one yet)
   valid_move=False

   # display the current state of the board
   show_board(board)
   
   # prompt for the player's move
   playermove_string=input("Enter your move (1-9):")
   
   # debug print
   if debug_print:
       print("DEBUG: player enters", playermove_string)
       
   # remove heading/trailing blanks from what was entered
   playermove_string=playermove_string.strip()
   
   # test if the input is numeric
   if playermove_string.isdigit():
       # the input is a number; convert it to an integer
       playermove=int(playermove_string)
       
       # debug print
       if debug_print:
           print("DEBUG: player move is", playermove)
           
       # test that the integer is between 1 and 9
       if playermove>=1 and playermove<=9:
           # the number is between 1 and 9
           # make sure the chosen cell is not already taken
           if not board[playermove-1]==mark_empty:
               print("Sorry, square",playermove,"is already taken")
           else:
               # put the player move on the board
               board[playermove-1]=mark_player
               # the player made a valid move
               valid_move=True
               # if the player has won, end the game
               if check_for_win(board, mark_player):
                   print("You win!")
                   break
                
               # count the number of cells filled
               full_cells=full_cells+1
               
               # debug print
               if debug_print:
                   print("DEBUG: full_cells is", full_cells)
                   
               # if the board is full, end the game in a draw
               if full_cells == 9:
                   # exits the loop
                   break
                
       # the player entered a number outside the range 1-9
       else:
           print("Enter your move as a number between 1 and 9")
   # the player entered a non-integer
   else:
       print("Enter your move as a number between 1 and 9")

   # if the board is full, end the game in a draw (could be redundant)
   if full_cells == 9:
        break
    
   # if the player made a valid move, pick a move for the computer
   if valid_move: 
       
      # pick a move for the computer
      computermove=pick_computer_move(board, full_cells)

      # put the computer move on the board
      board[computermove-1]=mark_computer
      # increment full cell count
      full_cells=full_cells+1
      
      # debug print
      if debug_print:
          print("DEBUG: full_cells is", full_cells)

      # check to see if the computer has won the game
      if check_for_win(board, mark_computer):
          print("The computer wins")
          break
        
      # check to see if the board is full
      if full_cells == 9:
          break

###################################   
# CLEANUP, TERMINATION, AND EXIT SECTION
###################################

# Display the board to illustrate the win.
# Note: you could add a show_winning_board()
# function to change winning
# tuple characters to asterisks.

show_board(board)
    
# when we get here, somebody won or the game is a draw
print("The game is over.")
ans = str(input("Push enter to exit: "))
